from .area import triangle_area_squared
from .point_in_polygon import point_in_polygon

from collections import defaultdict
from functools import lru_cache
from mathutils import Vector

class Portal:
	__slots__ = "source", "destination", "left", "right"

	def __init__(self, source, destination):
		self.source = source
		self.destination = destination

		self.left, self.right = self._generate_nodes()

	def _generate_nodes(self):
		source_position = self.source.position
		destination_position = self.destination.position
		first, second = [v for v in self.source.vertices if v in self.destination.vertices]

		side_first = triangle_area_squared(source_position, destination_position, first)
		side_second = triangle_area_squared(source_position, destination_position, second)

		mapping = {side_first: first, side_second: second}

		right = mapping[max(side_first, side_second)]
		left = mapping[min(side_first, side_second)]

		return left, right

		
class NavigationNode:

	def __init__(self, *vertices, neighbours=None):
		self.vertices = vertices
		self.position = sum(vertices, Vector()) / len(vertices)
		self.neighbours = neighbours or []

	@lru_cache()
	def get_portal_to(self, other):
		return Portal(self, other)

	def get_neighbours(self):
		return self.neighbours	

	def __contains__(self, point):
		return point_in_polygon(point, self.vertices)
	
	def get_common_vertices(self, other):
		return (v for v in self.vertices if v in other.vertices)


class BGEMesh:

	def __init__(self, obj):
		self._obj = obj
		self.polygons = self.build_nodes(obj.meshes[0])

	def build_nodes(self, mesh):
		points = self.get_approximate_points(mesh)
		
		nodes = []
		for polygon_id in range(mesh.numPolygons):
			polygon = mesh.getPolygon(polygon_id)
			positions = []
			for vertex_id in range(polygon.getNumVertex()):
				vertex_index = polygon.getVertexIndex(vertex_id)
				material_index = polygon.material_id

				point = points[material_index][vertex_index]
				positions.append(point)
			
			node = NavigationNode(*positions)
			nodes.append(node)
		
		found_neighbours = set()
		for node in nodes:
			for node_ in nodes:
				if node_ is node:
					continue
				if len(list(node.get_common_vertices(node_))) > 1:
					node.neighbours.append(node_)
					node_.neighbours.append(node)
		return nodes
		
	def get_approximate_points(self, mesh, epsilon = 0.001):	
		transform = self._obj.worldTransform
		material_points = {m_index: {v_index: (transform * mesh.getVertex(m_index, v_index).XYZ)
							for v_index in range(mesh.getVertexArrayLength(m_index))}
							for m_index in range(len(mesh.materials))}
	
		material_unique = defaultdict(dict)

		for material, points in material_points.items():
			unique_points = material_unique[material]
			for point_id, point in points.items():
				if point_id in unique_points:
					continue

				# Choose one and associate all others with it
				unique_points[point_id] = point
				for point_id_, point_ in points.items():
					if (point_ - point).length_squared < epsilon:
						unique_points[point_id_] = point
		return material_unique